from clease.montecarlo.observers import MCObserver
from clease.calculator import Clease
from typing import List, Dict, Tuple
from ase.units import kB
import numpy as np


class SGCState(object):
    """
    Represent a thermodynamic state in the semi-grand-canonical ensemble.

    :param temp: Temperature in kelvin
    :param chem_pot: Chemical potentials of the form {c1_0: -0.2, c1_1: 0.3}.
        The function `clease.tools.species_chempot2eci` is useful to convert
        chemical potentials given for each species to chemical potentials
        for each singlet.
    """
    def __init__(self, temp: float, chem_pot: Dict[str, float]):
        self.temp = temp
        self.chem_pot = chem_pot
        self.singlets = {k: 0.0 for k in self.chem_pot.keys()}
        self.avg_weight = 0.0

    def reset(self):
        self.singlets = {k: 0.0 for k in self.chem_pot.keys()}
        self.avg_weight = 0.0

    @property
    def prefix(self) -> str:
        """
        Construct a prefix based on the chemical potentials and the temperature
        """
        prefix = f"{int(self.temp)}K_"
        prefix += '_'.join(f"{k}{sign_indicator(v)}{int(1000.0*abs(v))}"
                           for k, v in self.chem_pot.items())
        return prefix


def sign_indicator(v: float) -> str:
    """
    Return minus if v is negative and plus if v is positive
    """
    if v < 0.0:
        return "minus"
    return "plus"


class MultiStateSGCConcObserver(MCObserver):
    """
    Observer that tracks the concentration at severl different temperatures
    and/or chemical potentials. The observer utilizes the following results.
    Let A be an observable, beta 1/kT, mu the chemical potential and n the
    number of atoms of one of the species in a binary alloy. The average value
    of the observable is given by

            sum_{conf} Aexp(beta*mu*n-beta*E)
    <A> = --------------------------------
                    Z(beta, mu)

    where Z is the partition function. At a different chemical potential mu'
    and inverse temperature beta',

             sum_{conf} Aexp(beta'*mu'*n-beta'*E)
    <A>' = ----------------------------------------
                    Z(beta', mu')

    After some algabraic manipulation one arrives at

              <Aexp((beta'*mu' - beta*mu)*n-(beta-beta')*E)>
    <A>' = ---------------------------------------------------
              <exp((beta'*mu' - beta*mu)*n-(beta-beta')*E)>

    where the averages should be taken at inverse temperature beta and chemical
    potential mu. It should be noted that the predicted value will not be
    accurate if mu' or beta' is very different from the reference values mu
    and beta.

    :param ref_state: Reference state
    :param thermo_states: List of SGCStates where the concentration should
        be tracked
    :param calc: Reference to the calculator attached to the atoms object
        used in the Monte Carlo simulation
    """
    def __init__(self, ref_state: SGCState, thermo_states: List[SGCState],
                 calc: Clease):
        self.thermo_states = thermo_states
        self.name = "multi_state_sgc_conc_observer"
        self.calc = calc
        self.ref_state = ref_state

    def reset(self) -> None:
        """
        Resets the observers to its initial state
        """
        for state in self.thermo_states:
            state.reset()

    def __call__(self, system_changes: List[Tuple[int, str, str]]):
        """
        Observers are called after system update, thus the calcualtor already
        reflects the changes. system_changes is therefore not used.
        """
        beta1 = 1.0/(kB*self.ref_state.temp)
        for state in self.thermo_states:
            cf = self.calc.get_cf()
            beta2 = 1.0/(kB*state.temp)
            betadE = sum((beta2*state.chem_pot[k] -
                          beta1*self.ref_state.chem_pot[k]) *
                         cf[k] for k in state.chem_pot.keys())
            E = self.calc.energy
            weight = np.exp((beta1 - beta2)*E + betadE)
            for k in state.chem_pot.keys():
                state.singlets[k] += weight*cf[k]
                state.avg_weight += weight

    def get_averages(self) -> Dict[str, float]:
        """
        Return a dictionary with the calculated averages
        """
        averages = {}
        for state in self.thermo_states:
            for k, v in state.singlets.items():
                name = state.prefix + f"_singlet_{k}"
                averages[name] = v/state.avg_weight
        return averages



import numpy as np
from typing import Union, Tuple, Optional
from ase import Atoms
from ase.build import surface, make_supercell
from ase.geometry import get_layers
from clease import (
    ClusterExpansionSettings, Concentration
)


def CESlab(conventional_cell: Union[Atoms, str],
           miller: Tuple[int], concentration: Concentration,
           size: Optional[Tuple[int]] = (1, 1, 1),
           max_cluster_size: Optional[int] = 4,
           max_cluster_dia: Optional[Tuple[float]] = (5.0, 5.0, 5.0),
           supercell_factor: Optional[int] = 27,
           db_name: Optional[str] = 'clease.db'
           ) -> ClusterExpansionSettings:
    """
    :param conventional_cell:
        Bulk lattice structure. Note that the unit-cell
        must be the conventional cell - not the primitive cell. One can also
        give the chemical symbol as a string, in which case the correct bulk
        lattice will be generated automatically.

    :param miller:
        Surface normal in Miller indices (h,k,l).

    :param concentration:
        Class for restricting the concentrations

    :param size:
        Size of the simulations cell. The third number represents the number of
        layers. The two first are repetitions of the in-plane unit vectors

    :param max_cluster_size:
        Maximum number of atoms in cluster

    :param max_cluster_dia:
        Maximum cluster diameters

    :param supercell_factor:
        Maximum number of unit cells that can be used to create a supercell

    :param db_name:
        Name of the database where structures should be placed
    """

    for b in concentration.basis_elements:
        if 'X' not in b:
            raise ValueError("Slab calculation requires that X is present "
                             "in all basis")

    prim = get_prim_slab_cell(conventional_cell, miller)

    # Slab should always have one cell vector along the z-axis
    settings = ClusterExpansionSettings(
        prim, concentration, size=size,
        max_cluster_dia=max_cluster_dia,
        max_cluster_size=max_cluster_size,
        supercell_factor=supercell_factor,
        db_name=db_name
    )

    dict_rep = conventional_cell.todict()
    for k, v in dict_rep.items():
        if isinstance(v, np.ndarray):
            dict_rep[k] = v.tolist()

    settings.kwargs.update(
        {
            'factory': 'CESlab',
            'miller': miller,
            'conventional_cell': dict_rep,
            'size': size
        }
    )
    return settings


def get_prim_slab_cell(conventional_cell: Union[Atoms, str],
                       miller: Tuple[int]) -> Atoms:
    """
    Returns the primitive cell used for slab CE

    :param conventional_cell:
        Bulk lattice structure. Note that the unit cell must be a conventional
        cell - not the primitive cell. One can also give the chemical symbol
        as a string, in which case the correct bulk lattice will be generated
        automatically.

    :param miller:
        Surface normal in Miller indices (h,k,l).
    """
    prim = surface(conventional_cell, miller, 1, periodic=True)
    return prim


def add_vacuum_layers(atoms: Atoms, prim: Atoms, thickness: float) -> Atoms:
    """
    Add vacuum layers to the slab

    :param atoms:
        ASE Atoms object representing the slab

    :param thickness:
        Thickness of the vacuum layer
    """
    # construct one layer of vacuum filled with vacancies
    prim_xy = prim.cell[:3]
    atoms_xy = atoms.cell[:3]
    P = atoms_xy.dot(np.linalg.inv(prim_xy))[:2].astype(int)
    P = np.vstack((P, [0, 0, 1]))
    vacuum = make_supercell(prim, P)
    for atom in vacuum:
        atom.symbol = 'X'

    dz = atoms.get_cell()[2, 2]
    vacuum.translate([0, 0, dz])

    one_layer_thickness = vacuum.get_cell()[2, 2]
    # upside-down division to get ceiling division
    num_vac_layers = int(-(-thickness//one_layer_thickness))
    vacuum *= (1, 1, num_vac_layers)

    atoms += vacuum
    atoms.cell[2, 2] += vacuum.cell[2, 2]
    return atoms


def remove_vacuum_layers(atoms: Atoms) -> Atoms:
    """
    Remove vacuum layers from the slab.

    :param atoms:
        ASE Atoms object representing the slab with vacuum
    """
    atoms = atoms.copy()
    tags, levels = get_layers(atoms, miller=(0, 0, 1))
    num_total_layers = np.amax(tags, axis=None) + 1
    vac_layers = []

    for layer in range(num_total_layers):
        symbols = [atoms[i].symbol for i,
                   tag in enumerate(tags) if tag == layer]
        if set(symbols) == {'X'}:
            vac_layers.append(layer)

    del atoms[[atoms[i].index for i,
               tag in enumerate(tags) if tag in vac_layers]]

    num_vac_layers = len(vac_layers)
    atoms.cell[2, 2] *= (num_total_layers - num_vac_layers)/num_total_layers

    return atoms



from clease.cleaselogger import CLEASELogger, LogVerbosity
_logger = CLEASELogger()


def set_verbosity(verbosity):
    _logger.verbosity = verbosity


def set_fd(fd):
    _logger.fd = fd


__version__ = '0.9.12'

from clease.template_filters import SkewnessFilter, EquivalentCellsFilter
from clease.template_filters import ValidConcentrationFilter
from clease.template_filters import DistanceBetweenFacetsFilter
from clease.template_filters import VolumeToSurfaceRatioFilter
from clease.template_filters import CellVectorDirectionFilter
from clease.template_filters import AngleFilter
from clease.atoms_manager import AtomsManager
from clease.settings import ClusterExpansionSettings
from clease.concentration import Concentration
from clease.settings_bulk import CEBulk, CECrystal, settingsFromJSON
from clease.settings_slab import CESlab
from clease.evaluate import Evaluate
from clease.corr_func import CorrFunction
from clease.new_struct import NewStructures
from clease.convexhull import ConvexHull
from clease.regression import LinearRegression, Tikhonov, Lasso
from clease.constrained_ridge import ConstrainedRidge
from clease.physical_ridge import PhysicalRidge
from clease.generalized_ridge_regression import (
    GeneralizedRidgeRegression, EigenDecomposition
)
from clease.ga_fit import GAFit
from clease.bayesian_compressive_sensing import BayesianCompressiveSensing
from clease.sequential_cluster_ridge import SequentialClusterRidge
from clease.cluster_info_mapper import ClusterInfoMapper
from clease.cluster_manager import ClusterManager
from clease.data_manager import (
    DataManager, CorrFuncEnergyDataManager, CorrFuncVolumeDataManager
)
from clease.structure_mapper import StructureMapper

__all__ = ['CEBulk', 'CECrystal', 'Concentration', 'CorrFunction',
           'NewStructures', 'NewStructures', 'Evaluate', 'AtomsManager',
           'ConvexHull', 'LinearRegression', 'Tikhonov', 'Lasso',
           'GAFit', 'BayesianCompressiveSensing', 'LogVerbosity',
           'SkewnessFilter', 'DataManager', 'CorrFuncEnergyDataManager',
           'CorrFuncVolumeDataManager', 'EquivalentCellsFilter',
           'ValidConcentrationFilter', 'DistanceBetweenFacetsFilter',
           'CellVectorDirectionFilter', 'AngleFilter',
           'ClusterExpansionSettings', 'settingsFromJSON', 'CESlab',
           'PhysicalRidge', 'SequentialClusterRidge', 'ClusterInfoMapper',
           'ClusterManager', 'VolumeToSurfaceRatioFilter', 'StructureMapper',
           'ConstrainedRidge', 'GeneralizedRidgeRegression',
           'EigenDecomposition']
